/*
 *   Copyright 2008 Ian McIntosh <ian@openanswers.org>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <stddef.h>
#include <math.h>
#include <stdarg.h>

//
// Joystick Support via SDL
//
#include <SDL/SDL.h>

//
// liblo OpenSoundControl (http://liblo.sourceforge.net/)
//
#include <lo/lo.h>

#define APPLICATION_NAME ("Joystick to OSC")
#define APPLICATION_VERSION ("0.1")
#define APPLICATION_COPYRIGHT ("Copyright 2008 Ian McIntosh")

#define UDP_PORT ("10007") // LOOOZ :D
#define UDP_ADDRESS NULL //("255.255.255.255")

typedef int TBoolean;
#define TRUE (1)
#define FALSE (0)

typedef struct
{
	int min;
	int max;
} TLimits;

typedef struct
{
	int axis_count;
	int *axis_values;
	TLimits *axis_limits;

	int button_count;
	TBoolean *button_values;

	SDL_Joystick* sdl_joystick;
} TJoystick;

float scale_and_expand_limits(int value, TLimits* limits)
{
	if(value < limits->min)
		limits->min = value;

	if(value > limits->max)
		limits->max = value;

	int range = (limits->max - limits->min);

	if(range == 0)
		return 0.0;

	return (float)(value - limits->min) / (float)range;
}

#define ADDRESS_BUFFER_SIZE 1000
char address_buffer[ADDRESS_BUFFER_SIZE + 1];

int main(int argc, char *argv[])
{
	printf("%s v%s %s\n", APPLICATION_NAME, APPLICATION_VERSION, APPLICATION_COPYRIGHT);

	if(SDL_Init(SDL_INIT_JOYSTICK) < 0 ) {
		fprintf(stderr, "Couldn't initialize SDL: %s\n", SDL_GetError());
		exit(1);
	}

	TJoystick* joysticks = malloc(sizeof(TJoystick) * SDL_NumJoysticks());

	int joystick_index;
	int button_index;
	int axis_index;

	for(joystick_index=0; joystick_index<SDL_NumJoysticks() ; joystick_index++) {
		TJoystick* joystick = (TJoystick*)&joysticks[joystick_index];

		joystick->sdl_joystick = SDL_JoystickOpen(joystick_index);

		// Add axes
		joystick->axis_count = SDL_JoystickNumAxes(joystick->sdl_joystick);
		joystick->axis_values = calloc(joystick->axis_count, sizeof(int));
		joystick->axis_limits = calloc(joystick->axis_count, sizeof(TLimits));		// 0 for min and max is fine for joystick axes

		// Add buttons
		joystick->button_count = SDL_JoystickNumButtons(joystick->sdl_joystick);
		joystick->button_values = calloc(joystick->button_count, sizeof(TBoolean));

		printf("Opened Input Device '%s'\n", SDL_JoystickName(joystick_index));
	}

	// Where to send OpenSoundControl messages
	lo_address t = lo_address_new(UDP_ADDRESS, UDP_PORT);

	printf("Ctrl-C to quit...\n");
	while (1) {
		SDL_JoystickUpdate();		// Updates state of all joysticks

		for(joystick_index=0; joystick_index<SDL_NumJoysticks() ; joystick_index++) {
			TJoystick* joystick = (TJoystick*)&joysticks[joystick_index];

			//
			// Update buttons
			//
			for(button_index=0; button_index<joystick->button_count ; button_index++) {
				if(SDL_JoystickGetButton(joystick->sdl_joystick, button_index) == SDL_PRESSED) {
					// Newly pressed?
					if(joystick->button_values[button_index] == FALSE) {
						joystick->button_values[button_index] = TRUE;

						snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "Joystick %02d / Button %02d", joystick_index + 1, button_index + 1);
						lo_send(t, address_buffer, "i", 1);
					}
				}
				else {
					// Newly released?
					if(joystick->button_values[button_index] == TRUE) {
						joystick->button_values[button_index] = FALSE;

						snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "Joystick %02d / Button %02d", joystick_index + 1, button_index + 1);
						lo_send(t, address_buffer, "i", 0);
					}
				}
			}

			//
			// Update axes
			//
			for(axis_index=0; axis_index<joystick->axis_count ; axis_index++) {
				int axis_value = SDL_JoystickGetAxis(joystick->sdl_joystick, axis_index);
				// Moved?
				if(axis_value != joystick->axis_values[axis_index]) {
					float value = scale_and_expand_limits(axis_value, &(joystick->axis_limits[axis_index]));

					snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "Joystick %02d / Axis %02d", joystick_index + 1, axis_index + 1);
					lo_send(t, address_buffer, "f", value);

					joystick->axis_values[axis_index] = axis_value;		// NOTE: this saves the raw integer axis_value
				}
			}
		}
		usleep(5 * 1000);
	}
}
