#!/usr/bin/env ruby

 ###############################################################################
 #  Copyright 2008 Ian McIntosh <ian@openanswers.org>
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
 #  the Free Software Foundation; either version 2 of the License, or
 #  (at your option) any later version.
 #
 #  This program is distributed in the hope that it will be useful,
 #  but WITHOUT ANY WARRANTY; without even the implied warranty of
 #  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 #  GNU Library General Public License for more details.
 #
 #  You should have received a copy of the GNU General Public License
 #  along with this program; if not, write to the Free Software
 #  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 ###############################################################################

Dir.chdir(File.dirname(__FILE__))	# So that this file can be run from anywhere
$LOAD_PATH.unshift('./utils').unshift('.')

###################################################################
# Constants
###################################################################
APP_NAME			= 'Luz Input Editor'
APP_COPYRIGHT	= "Copyright (c) #{Time.now.year} Ian McIntosh"
APP_AUTHORS 	= ['Ian McIntosh <ian@openanswers.org>']
APP_VERSION		= 0.1
WINDOW_TITLE	= "Luz Input Editor"

require 'openal'

require 'reloadable_require'
require 'addons_ruby', 'method_piping', 'boolean_accessor'

# Application basics
require 'constants', 'sdl', 'opengl', 'addons_gl', 'drawing'

require 'gl_curve_editor', 'gl_progress_bar', 'gl_event_editor'

def fail(msg)
	puts msg
	exit
end

class LuzInputEditor
	include Drawing

	attr_accessor :width, :height, :fullscreen, :bits_per_pixel, :frames_per_second

	def initialize
		@width, @height, @bits_per_pixel = 0, 0, 0		# NOTE: setting bpp to 0 means "current" in SDL
		@fullscreen = true

		# Frame rate should, ideally, match that of LCD, projector, etc.
		# TODO: add an option for syncing to output device refresh rate, as a way to limit refresh rate.
		@frames_per_second = 70

		@zoom_range = 0.1

		@widgets = []
	end

	def create
		#
		# Init OpenAL
		#
		ALUT.init
		@source = AL::Source.new

		#
		# Init SDL
		#
		SDL.init(SDL::INIT_VIDEO | SDL::INIT_TIMER)
		puts "Using SDL version #{SDL::VERSION}"

		@sdl_video_mode_flags = SDL::HWSURFACE | SDL::OPENGL
		@sdl_video_mode_flags |= SDL::FULLSCREEN if @fullscreen

		@screen = SDL.set_video_mode(@width, @height, @bits_per_pixel, @sdl_video_mode_flags)
		@width, @height = @screen.w, @screen.h

		# save useful info if we were using bpp=0 ("current")
		@bits_per_pixel = @screen.bpp if @bits_per_pixel == 0
		puts "Running at #{@screen.w}x#{@screen.h} @ #{@bits_per_pixel}bpp, #{@frames_per_second}fps (max)"

		SDL::WM.set_caption(WINDOW_TITLE, '')
		SDL::Key.enable_key_repeat(100, 100)

		# setup_opengl
		GL.Enable(GL::BLEND)
		GL.BlendFunc(GL::SRC_ALPHA, GL::ONE_MINUS_SRC_ALPHA)
		GL.Disable(GL::DEPTH_TEST)

		GL.Viewport(0, 0, @width, @height)

		GL.EnableClientState(GL::VERTEX_ARRAY)

		GL.MatrixMode(GL::PROJECTION)
		GL.LoadIdentity
		GLU.Perspective(90.0, @width/@height, 0.1, 100.0)
		GLU.LookAt(
			0.0, 0.0, 0.5,		# position
			0.0, 0.0, 0.0,		# looking at
			0.0, 1.0, 0.0			# up
		)

		GL.MatrixMode(GL::MODELVIEW)
		GL.LoadIdentity

		# Rendering should fill screen/window completely, regardless of dimensions
		GL.Scale((@width > @height) ? @width / @height : 1.0, (@height > @width) ? @height / @width : 1.0, 1.0)
	end

	EDITOR_SPACING = 0.04
	CURVE_AND_EVENT_SPACING = 0.01

	def load(audio_filepath)
		@audio_filepath = audio_filepath
		@buffer = AL::Buffer.load_from_file(@audio_filepath)
		fail("failed to load #{@audio_filepath}") unless @buffer

		@buffer_size = @buffer.size
		@buffer_sample_multiplier = (@buffer.channels * (@buffer.bits / 8))

		@buffer_length = @buffer_size.to_f / (@buffer.frequency * (@buffer.channels * (@buffer.bits / 8)))

		@source.attach(@buffer)

		#
		# Load input files
		#
		@curves = {}
		Dir.glob("#{@audio_filepath}.*").each { |path|
			name = File.extname(path).without_prefix('.')
			puts "Loading #{name}..."
			@curves[name] ||= [[], []]

			File.open(path, 'r') { |f|
				f.each_line { |line|
					columns = line.split(',').collect { |c| c.strip }
					time = columns[0].to_f / @buffer_length
					if columns[2] == 'true' || columns[2] == 'false'
						@curves[name].first << GLCurveEditor::Sample.new(time, (columns[2] == 'true'))
					else
						@curves[name].last << GLCurveEditor::Sample.new(time, columns[2].to_f)
					end
				}
			}
		}
		puts "Loaded #{@curves.size} curves."

		#
		# Create GUI from input files
		#
		y = 0.5

		#
		# add progress bar
		#
		height = 0.04
		y -= height/2.0
		@progress_bar = GLProgressBar.new(0.0, y, 1.0, height)
		@widgets << @progress_bar
		y -= height/2.0

		#
		# add curve editor
		#
		y -= EDITOR_SPACING
		height = ((y + 0.5) / @curves.size)
		height -= EDITOR_SPACING
		height -= CURVE_AND_EVENT_SPACING

		@curves.each { |key, value|
			y -= (height * 0.7)/2.0
			@widgets << GLCurveEditor.new(0.0, y, 1.0, height * 0.7)
			@widgets.last.samples = value.last
			@widgets.last.name = key
			y -= (height * 0.7)/2.0

			y -= CURVE_AND_EVENT_SPACING

			y -= (height * 0.3)/2.0
			@widgets << GLEventEditor.new(0.0, y, 1.0, height * 0.3)
			@widgets.last.samples = value.first
			@widgets.last.name = key
			y -= (height * 0.3)/2.0

			y -= EDITOR_SPACING

			# other options
		}

		#
		# progress bar click-and-drag-to-seek support
		#
		@progress_bar.on_seek { |progress|
			@source.sec_offset = (progress * @buffer_length)
		}
	end

	def save
		puts 'saving...'
		@curves.each { |key, value|
			filename = @audio_filepath + '.' + key
			File.open(filename, 'w') { |f|
				# both float and boolean
				last_val, last_time = nil, nil
				bools = value[0].delete_if { |sample| (sample.value == last_val || sample.time == last_time) ? (true) : (last_val, last_time=sample.value, sample.time ; false) }
				last_val, last_time = nil, nil
				floats = value[1].delete_if { |sample| (sample.value == last_val || sample.time == last_time) ? (true) : (last_val, last_time=sample.value, sample.time ; false) }

				samples = (bools + floats).sort_by { |sample| sample.time }
				samples.each { |sample|
					time = sample.time * @buffer_length
					f.write(sprintf("%s, %s, %s\n", time, key, sample.value.to_s))
				}
			}
		}
	end

	def percentage
		((@source.sample_offset.to_f * @buffer_sample_multiplier) / @buffer_size)
	end

	boolean_accessor :finished

	def run
		ms_per_frame = (1000 / @frames_per_second)

		# get it to pause state so we can fast-forward and rewrind
		@source.play
		@source.pause

		while not finished?
			frame_start_ms = SDL.getTicks

			handle_input_events

			#
			# Render then sleep until time for next frame
			#
			update(percentage)

			draw(percentage)
			SDL.GL_swap_buffers

			# allow GC when paused
			GC.enable unless @source.playing?

			frame_time_ms = SDL.getTicks - frame_start_ms

			SDL.delay(ms_per_frame - frame_time_ms) if frame_time_ms < ms_per_frame
		end
		SDL.quit

		puts "~#{1.0 / @last_draw_time} fps" if @last_draw_time
		save
	end

	def handle_input_events
		while event = SDL::Event2.poll
		case event
			# Mouse input
			when SDL::Event2::MouseMotion
				@widgets.each { |widget| widget.cursor_move(percentage, (event.x / (@width - 1).to_f)-0.5, (1.0 - (event.y / (@height - 1).to_f))-0.5) }
			when SDL::Event2::MouseButtonDown
				@widgets.each { |widget| widget.button_down(percentage, event.button, (event.x / (@width - 1).to_f)-0.5, (1.0 - (event.y / (@height - 1).to_f))-0.5) }
			when SDL::Event2::MouseButtonUp
				@widgets.each { |widget| widget.button_up(percentage, event.button, (event.x / (@width - 1).to_f)-0.5, (1.0 - (event.y / (@height - 1).to_f))-0.5) }

			# Keyboard input
			when SDL::Event2::KeyDown
				case event.sym
				when SDL::Key::ESCAPE
					self.finished = true
				when SDL::Key::SPACE
					toggle_paused!
				when SDL::Key::LEFT
					@source.sec_offset = [@source.sec_offset - (1 + 10.0 * (@zoom_range ** 2)), 0].max
				when SDL::Key::RIGHT
					@source.sec_offset += (1 + (10.0 * (@zoom_range ** 2)))
				when SDL::Key::EQUALS
					@zoom_range = (@zoom_range - 0.04).clamp(0.02, 1.0)
				when SDL::Key::MINUS
					@zoom_range = (@zoom_range + 0.04).clamp(0.02, 1.0)
				end

			when SDL::Event2::KeyUp

			# Window closing, etc.
			when SDL::Event2::Quit
				self.finished = true
			end
		end
	end

	def update(percentage)
		@widgets.each { |widget| widget.update(percentage) }
	end

	def draw(percentage)
		t=Time.new
		clear_screen([0.0, 0.0, 0.0, 0.0])

		@widgets.each { |widget| widget.draw(percentage, @zoom_range) }

		@last_draw_time = Time.new-t
	end

	def toggle_paused!
		if @source.playing?
			@source.pause
		else
			GC.disable
			@source.play
		end
	end
end

$application = LuzInputEditor.new

# Last argument is project name
unless (audio_path = ARGV.pop)
	puts 'missing audio file'
	exit
end

#
# Parse command-line options
#
require 'optparse'

options = OptionParser.new do |opts|
	opts.banner = "Usage: luz_input_editor.rb [options] audio.wav"

	opts.on("--width <width>", Integer, "Resolution width (eg. 800)") do |w|
		$application.width = w
	end
	opts.on("--height <height>", Integer, "Resolution height (eg. 600)") do |h|
		$application.height = h
	end
	opts.on("--bits-per-pixel <bpp>", Integer, "Bits per pixel (8, 16, 24, 32)") do |bpp|
		$application.bits_per_pixel = bpp
	end
	opts.on("--frames-per-second <fps>", Integer, "Target FPS") do |fps|
		$application.frames_per_second = fps.to_i
	end
	opts.on("--fullscreen", "Fullscreen") do
		$application.fullscreen = true
	end
	opts.on("--window", "Window") do
		$application.fullscreen = false
	end
end

options.parse!

$application.create
$application.load(audio_path)
$application.run
