#!/usr/bin/env ruby

 ###############################################################################
 #  Copyright 2009 Ian McIntosh <ian@openanswers.org>
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
 #  the Free Software Foundation; either version 2 of the License, or
 #  (at your option) any later version.
 #
 #  This program is distributed in the hope that it will be useful,
 #  but WITHOUT ANY WARRANTY; without even the implied warranty of
 #  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 #  GNU Library General Public License for more details.
 #
 #  You should have received a copy of the GNU General Public License
 #  along with this program; if not, write to the Free Software
 #  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 ###############################################################################

Dir.chdir(File.dirname(__FILE__))	# So that this file can be run from anywhere
$LOAD_PATH.unshift('./utils').unshift('.')

###################################################################
# Constants
###################################################################
APP_NAME			= 'Luz Playback'
APP_COPYRIGHT	= "Copyright (c) #{Time.now.year} Ian McIntosh"
APP_AUTHORS 	= ['Ian McIntosh <ian@openanswers.org>']
APP_VERSION		= 0.1

SLEEP_TIME = 0.005		# ~200 FPS

Sample = Struct.new(:time, :name, :value)

require 'socket'
class IPSocket
	def set_reuse_address_flag
		setsockopt(Socket::SOL_SOCKET, Socket::SO_REUSEADDR, 1)
		self
	end
end

def fail(msg)
	puts msg
	exit
end

# Last argument is project name
fail('syntax: ./luz_playback.rb music.wav') unless ARGV.size == 1
audio_filepath = ARGV.shift
input_filepath = audio_filepath + '.luz.inputs'

require 'constants'
require 'osc'

#
# Init OpenAL
#
require 'openal'

class LuzPlayback
	attr_reader :samples

	def initialize
		ALUT.init
		puts "#{AL.string(AL::RENDERER)} #{AL.string(AL::VERSION)}"

		# Source
		@source = AL::Source.new
		@source.position = [0.0, 0.0, 0.0]
		@source.source_relative = true

		@samples = []

		@socket = UDPSocket.new.set_reuse_address_flag

#		@socket.connect("225.4.5.6", MESSAGE_BUS_PORT)

		@socket.setsockopt(Socket::SOL_SOCKET, Socket::SO_BROADCAST, true)
		@socket.connect("255.255.255.255", MESSAGE_BUS_PORT)
	end

	def load(audio_filepath)
		puts "Loading audio..."

		@buffer = AL::Buffer.load_from_file(audio_filepath)
		fail("failed to load #{audio_filepath}") unless @buffer
		@source.attach(@buffer)

		# helpful precalcations
		@buffer_size = @buffer.size
		@buffer_sample_multiplier = (@buffer.channels * (@buffer.bits / 8))		# multiply a sample offset by this to get the overall progress through the buffer

		@buffer_length = @buffer_size.to_f / (@buffer.frequency * (@buffer.channels * (@buffer.bits / 8)))
		puts "Loaded #{audio_filepath} (#{@buffer_length} seconds)..."

		puts "Loading inputs..."
		Dir.glob("#{audio_filepath}.*").each { |path|
			File.open(path, 'r') { |f|
				start = @samples.size
				f.each_line { |line|
					columns = line.split(',').collect { |c| c.strip }
					@samples << Sample.new(columns[0].to_f, columns[1], {'true' => 1, 'false' => 0}[columns[2]] || columns[2].to_f)
				}
				puts "Loaded #{path} with #{@samples.size - start} samples..."
			}
		}
		puts 'Sorting...'
		@samples = @samples.sort_by { |s| s.time }
	end

	def percentage
		((@source.sample_offset.to_f * @buffer_sample_multiplier) / @buffer_size)
	end

	def run
		@source.play
		@start_time = Time.new.to_f
	end

	def update
		@index ||= 0

		offset = percentage * @buffer_length

		while (sample=@samples[@index]) and sample.time < offset
			send_sample(sample)
			@index += 1
		end
	end

	def send_sample(sample)
		m = OSC::Message.new(sample.name, nil, sample.value)
		@socket.send(m.encode, 0)
	end
end

app = LuzPlayback.new
app.load(audio_filepath)

#
# Begin!
#
GC.start
GC.disable
app.run

begin
	while true
		app.update
		sleep(SLEEP_TIME)
	end
rescue Interrupt
	puts # blank line for a ^C etc. in the terminal
	puts 'Interrupted.'
end

=begin
require 'sdl'
 ... the SDL version ...
SDL::init(SDL::INIT_AUDIO)
SDL::Mixer.open(44100, SDL::Mixer::DEFAULT_FORMAT, 2, 512)
wave = SDL::Mixer::Wave.load(filename)
SDL::Mixer.play_channel(0, wave, 0)
while SDL::Mixer.play?(0)
  sleep 1
end
=end
