#!/usr/bin/env ruby

 ###############################################################################
 #  Copyright 2009 Ian McIntosh <ian@openanswers.org>
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
 #  the Free Software Foundation; either version 2 of the License, or
 #  (at your option) any later version.
 #
 #  This program is distributed in the hope that it will be useful,
 #  but WITHOUT ANY WARRANTY; without even the implied warranty of
 #  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 #  GNU Library General Public License for more details.
 #
 #  You should have received a copy of the GNU General Public License
 #  along with this program; if not, write to the Free Software
 #  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 ###############################################################################

Dir.chdir(File.dirname(__FILE__))	# So that this file can be run from anywhere
$LOAD_PATH.unshift('./utils').unshift('.')

###################################################################
# Constants
###################################################################
APP_NAME			= 'Luz Recorder'
APP_COPYRIGHT	= "Copyright (c) #{Time.now.year} Ian McIntosh"
APP_AUTHORS 	= ['Ian McIntosh <ian@openanswers.org>']
APP_VERSION		= 0.1

SLEEP_TIME = 0.005		# ~200 FPS

Sample = Struct.new(:time, :name, :value)

def fail(msg)
	puts msg
	exit
end

# Last argument is project name
fail('syntax: ./luz_recorder.rb music.wav') unless ARGV.size == 1
audio_filepath = ARGV.shift
output_filepath = audio_filepath + '.luz.inputs'

#
# Init message bus
#
require 'message_bus'
require 'constants'

#
# Init OpenAL
#
require 'openal'

class LuzRecorder
	attr_reader :samples

	def initialize
		ALUT.init
		puts "#{AL.string(AL::RENDERER)} #{AL.string(AL::VERSION)}"

		# Source
		@source = AL::Source.new
		@source.position = [0.0, 0.0, 0.0]
		@source.source_relative = true

		@samples = []
		@offset_in_seconds = 0.0

		@message_bus = MessageBus.new
		@message_bus.on_button_down { |button_name| @samples << Sample.new(@offset_in_seconds, button_name.dup, true) }
		@message_bus.on_button_up { |button_name| @samples << Sample.new(@offset_in_seconds, button_name.dup, false) }
		@message_bus.on_slider_change { |slider_name, value| @samples << Sample.new(@offset_in_seconds, slider_name.dup, value) }
	end

	def load(audio_filepath)
		@buffer = AL::Buffer.load_from_file(audio_filepath)
		fail("failed to load #{audio_filepath}") unless @buffer
		@source.attach(@buffer)
	end

	def run
		@message_bus.listen(MESSAGE_BUS_IP, MESSAGE_BUS_PORT)
		@source.play
		@start_time = Time.new.to_f
	end

	def update
		return false unless @source.playing?

		@offset_in_seconds = Time.new.to_f - @start_time
		@message_bus.update

		true
	end

	def save_samples_to_file(output_filepath)
		if @samples.empty?
			puts "Not saving (0 samples received)."
		else
			puts "Saving #{@samples.size} samples to #{output_filepath}..."

			File.open(output_filepath, 'w') { |f|
				@samples.each_with_index { |s, index|
					@source.gain = 1.0 - (index.to_f / @samples.size.to_f)
					f.write("#{s.time}, #{s.name}, #{s.value}\n")
				}
			}
			puts "Saved."
		end
	end
end

app = LuzRecorder.new
app.load(audio_filepath)

#
# Wait for it...
#
puts 'Begin in...'
sleep 1
for n in 0...3
	puts 3-n
	sleep 1
end

puts 'Here we go!'

#
# Begin!
#
GC.start
GC.disable
app.run

begin
	while true
		break unless app.update
		sleep(SLEEP_TIME)
	end
rescue Interrupt
	puts # blank line for a ^C etc. in the terminal
	puts 'Interrupted.'
end

app.save_samples_to_file(output_filepath)

=begin
require 'sdl'
 ... the SDL version ...
SDL::init(SDL::INIT_AUDIO)
SDL::Mixer.open(44100, SDL::Mixer::DEFAULT_FORMAT, 2, 512)
wave = SDL::Mixer::Wave.load(filename)
SDL::Mixer.play_channel(0, wave, 0)
while SDL::Mixer.play?(0)
  sleep 1
end
=end
