/*
 *   Copyright 2008 Ian McIntosh <ian@openanswers.org>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <stddef.h>
#include <math.h>
#include <stdarg.h>

//
// Portmidi
//
#include <portmidi.h>
#include <porttime.h>

//
// liblo OpenSoundControl (http://liblo.sourceforge.net/)
//
#include <lo/lo.h>

#define APPLICATION_NAME ("MIDI to OSC")
#define APPLICATION_VERSION ("0.1")
#define APPLICATION_COPYRIGHT ("Copyright 2008 Ian McIntosh")

#define UDP_PORT ("10007") // LOOOZ :D
#define UDP_ADDRESS NULL //("255.255.255.255")

#define MAX_INPUTS (100)

#define TIME_PROC ((long (*)(void *)) Pt_Time)
#define INPUT_BUFFER_SIZE (100)

#define ADDRESS_BUFFER_SIZE 1000
char address_buffer[ADDRESS_BUFFER_SIZE + 1];

int main(int argc, char *argv[])
{
	char address_buffer[ADDRESS_BUFFER_SIZE + 1];
	printf("%s v%s %s\n", APPLICATION_NAME, APPLICATION_VERSION, APPLICATION_COPYRIGHT);

	lo_address t = lo_address_new(UDP_ADDRESS, UDP_PORT);


	Pm_Initialize();

	PortMidiStream* inputs[MAX_INPUTS] = {0};
	int num_inputs = 0;

	int i;
	int nDevices = Pm_CountDevices();

	for(i=0 ; i<nDevices ; i++) {
		PmDeviceInfo *info = Pm_GetDeviceInfo(i);
		if(info == NULL)
			continue;

		if(info->input == TRUE) {
			PortMidiStream* input = NULL;
			Pm_OpenInput(&input, i, NULL, INPUT_BUFFER_SIZE, TIME_PROC, NULL);
			if(input) {
				printf("Opened Input Device '%s'\n", info->name);

				// Add to list
				inputs[num_inputs] = input;
				num_inputs++;
			}
		}
	}

	//
	// Loop and dispatch events
	//

	printf("Ctrl-C to quit...\n");
	while(1) {
		// Check one input after another
		for(i=0 ; i<num_inputs ; i++) {
			PortMidiStream* input = inputs[i];

			// Read all events
			PmEvent event;
			while(Pm_Poll(input) && (Pm_Read(input, &event, 1) == 1)) {
				int channel = (Pm_MessageStatus(event.message) & 0x0F) + 1;		// +1 so we send out channel messages in the range 1-16
				int type = (Pm_MessageStatus(event.message) & 0xF0) >> 4;
				int data_1 = Pm_MessageData1(event.message);
				int data_2 = Pm_MessageData2(event.message);

				// Debug output
				//if(data_1 != 0 || data_2 != 0) {
				//	printf("channel=%d, type=%d, one=%d, two=%d\n", channel, type, data_1, data_2);
				//}

				// Slider change
				if(type == 11) {
					snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "MIDI / Channel %02d / Slider %03d", channel, data_1);
					lo_send(t, address_buffer, "f", (float)data_2 / 127.0);

					if(data_2 == 127) {
						// Axiom25 buttons send this for presses
						snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "MIDI / Channel %02d / Button %03d", channel, data_1);
						lo_send(t, address_buffer, "i", 1);
					}
					else {		// usually 0 but accept anything
						snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "MIDI / Channel %02d / Button %03d", channel, data_1);
						lo_send(t, address_buffer, "i", 0);
					}
				}

				//
				// Type 14 is a spring-loaded Pitch Bend control
				//
				else if(type == 14) {
					snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "MIDI / Channel %02d / Pitch Bend", channel);
					lo_send(t, address_buffer, "f", (data_2 == 64) ? 0.5 : (float)data_2 / 127.0);	// HACK: 64 is the at-rest value on the M-Audio Axiom 25 (remove this if found to be uncommon)
				}

				//
				// Channel Aftertouch
				//
				else if(type == 13) {
					snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "MIDI / Channel %02d / Aftertouch", channel);
					lo_send(t, address_buffer, "f", (float)data_1 / 127.0);
				}

				else if(type == 9) {
					snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "MIDI / Channel %02d / Key %03d", channel, data_1);
					lo_send(t, address_buffer, "f", (float)data_2 / 127.0);

					// NOTE: many MIDI controllers (eg. M-Audio X-Session send buttons via Note messages)
					if(data_2 == 0) {
						// Note/Button off
						snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "MIDI / Channel %02d / Button %03d", channel, data_1);
						lo_send(t, address_buffer, "i", 0);
					}
					else {		// anything non-0 is an 'on'
						// Note on
						snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "MIDI / Channel %02d / Button %03d", channel, data_1);
						lo_send(t, address_buffer, "i", 1);
					}
				}
			}
		}
		usleep(4 * 1000);
	}
	// Never get here...
	//Pm_Terminate();
}
