/*
 *   Copyright 2008 Ian McIntosh <ian@openanswers.org>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>

// ALSA audio
#define ALSA_PCM_NEW_HW_PARAMS_API		// Use the newer ALSA API
#include <alsa/asoundlib.h>

// Aubio Audio Analysis
#include <aubio/aubio.h>
#include <aubio/aubioext.h>

//
// liblo OpenSoundControl (http://liblo.sourceforge.net/)
//
#include <lo/lo.h>
#define UDP_PORT ("10007") // LOOOZ :D

// energy, specdiff, hfc, complexdomain, phase
aubio_onsetdetection_type type_onset  = aubio_onset_kl;
smpl_t threshold                      = 5.0;		// 0.3
smpl_t silence                        = -50.0;	// -90.0
uint_t buffer_size                    = 2048;
uint_t overlap_size                   = 1024;

#define NUM_CHANNELS (1)

int main(int argc, char **argv)
{
	//
	// Init ALSA
	//
	int ret;
	int dir = 0;
	unsigned int rate = 44100;
	snd_pcm_uframes_t frames = 256;

	if(argc < 2) {
		printf("Usage: %s <alsa capture device>\n", argv[0]);
		return 1;
	}

	char* alsa_device_name = argv[1];

	snd_pcm_t *alsa_handle;
	ret = snd_pcm_open(&alsa_handle, alsa_device_name, SND_PCM_STREAM_CAPTURE, 0);
	if(ret < 0) {
		printf("unable to open pcm device '%s': %s\n", alsa_device_name, snd_strerror(ret));
		return 1;
	}

	// Get default params
	snd_pcm_hw_params_t *params;
	snd_pcm_hw_params_alloca(&params);
	snd_pcm_hw_params_any(alsa_handle, params);

	// Set desired params
	snd_pcm_hw_params_set_access(alsa_handle, params, SND_PCM_ACCESS_RW_INTERLEAVED);
	snd_pcm_hw_params_set_format(alsa_handle, params, SND_PCM_FORMAT_FLOAT); //SND_PCM_FORMAT_S16_LE);
	snd_pcm_hw_params_set_channels(alsa_handle, params, NUM_CHANNELS);
	snd_pcm_hw_params_set_rate_near(alsa_handle, params, &rate, &dir);
	snd_pcm_hw_params_set_period_size_near(alsa_handle, params, &frames, &dir);

	// Write the parameters to the driver
	ret = snd_pcm_hw_params(alsa_handle, params);
	if(ret < 0) {
		fprintf(stderr, "unable to set hw parameters: %s\n", snd_strerror(ret));
		exit(1);
	}
	snd_pcm_hw_params_get_period_size(params, &frames, &dir);
	snd_pcm_hw_params_get_rate(params, &rate, &dir);

	// Create a sample buffer large enough to hold one period
	char* sample_buffer = (char*)malloc(frames * NUM_CHANNELS * sizeof(float));

	//
	// Init Aubio
	//
	fvec_t* woodblock = new_fvec(buffer_size, 1);
	fvec_t* ibuf = new_fvec(overlap_size, NUM_CHANNELS);
	fvec_t* obuf = new_fvec(overlap_size, NUM_CHANNELS);
	fvec_t* out = new_fvec(2, NUM_CHANNELS);

	aubio_tempo_t* aubio_tempo_detector = new_aubio_tempo(type_onset, buffer_size, overlap_size, NUM_CHANNELS);

	printf("frames=%d, overlap_size=%d, length=%d, channels=%d\n", frames, overlap_size, ibuf->length, ibuf->channels);

	lo_address t = lo_address_new(NULL, UDP_PORT);

	//
	// Loop, read sound samples, pass to Aubio, and send off beats and junk via OSC
	//

	uint_t istactus = 0;
	//int num_frames = 0;
		unsigned int pos = 0;    // frames % dspblocksize

	while(1)
	{
		// Read some audio data
		ret = snd_pcm_readi(alsa_handle, sample_buffer, frames);
		if(ret == -EPIPE) {
			// EPIPE means overrun
			fprintf(stderr, "overrun occurred\n");
			snd_pcm_prepare(alsa_handle);
			continue;
		} else if(ret < 0) {
			fprintf(stderr, "error from read: %s\n", snd_strerror(ret));
			continue;
		} else if(ret != (int)frames) {
			fprintf(stderr, "short read, read %d frames\n", ret);
			continue;
		}

/* Buffer for real values
struct _fvec_t {
  uint_t length;   // length of buffer
  uint_t channels; // number of channels
  smpl_t **data;   // data array of size [length] * [channels]
};
*/
	ibuf->length = frames; //overlap_size;						// <==== or frames? or..?
	ibuf->channels = NUM_CHANNELS;
	ibuf->data = (smpl_t**)&sample_buffer;		// <========= is this right?

		// Aubio processing
		//aubio_sndfile_read(file_handle, overlap_size, ibuf);
		unsigned int i;

		for(i = 0 ; i < (unsigned)overlap_size ; i++) {
			// time for FFT?
			if(pos == overlap_size-1) {
				aubio_tempo(aubio_tempo_detector, ibuf, out);

				if(out->data[0][0]==1)
					istactus = 1;
				else
					istactus = 0;

				if (istactus) {
					for (pos = 0; pos < overlap_size; pos++) {
						obuf->data[0][pos] = woodblock->data[0][pos];
					}
				} else {
					for (pos = 0; pos < overlap_size; pos++) {
						obuf->data[0][pos] = 0.0;
					}
				}

				pos = -1; // so it will be zero for next loop
			}
			pos++;
		}

		if(istactus) {
			printf("beat...\n");
			lo_send(t, "Music Analysis / Beat", "i", 1);
			lo_send(t, "Music Analysis / Beat", "i", 0);
		}
	}
	del_aubio_tempo(aubio_tempo_detector);
	del_fvec(out);

	return 0;
}
