/*
 *   Copyright 2009 Ian McIntosh <ian@openanswers.org>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "spectrum-analyzer.h"
#include "utils.h"
#include "filter.h"
#include "audio-sampler.h"

#include <math.h>
#include <vector>

using namespace std;    // saves us typing std:: before vector

#include <GL/gl.h>	// Header File For The OpenGL32 Library
#include <GL/glu.h>	// Header File For The GLu32 Library

#include <SDL/SDL.h>

#define APPLICATION_NAME ("Luz Spectrum Analyzer")
#define APPLICATION_VERSION ("0.31")
#define APPLICATION_COPYRIGHT ("Copyright 2008 Ian McIntosh")

#define WINDOW_TITLE ("Luz Spectrum Analyzer - F12: fullscreen, Esc: minimize")

//
// liblo OpenSoundControl (http://liblo.sourceforge.net/)
//
#include <lo/lo.h>

#define UDP_PORT ("10007") // LOOOZ :D
#define UDP_ADDRESS NULL //("255.255.255.255")

#define ADDRESS_BUFFER_SIZE 1000
char address_buffer[ADDRESS_BUFFER_SIZE + 1];

//
// Utility
//
void render_unit_square()
{
	glVertex2f(-0.5, 0.5);
	glVertex2f( 0.5, 0.5);
	glVertex2f( 0.5, -0.5);
	glVertex2f(-0.5, -0.5);
}

//
// OpenGL Rendering
//
void initialize_OpenGL(int width, int height)
{
	// Default options
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glClearColor(0.05, 0.05, 0.05, 1.0);

	// Default view
	glViewport(0, 0, width, height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90.0f,(GLfloat)width/(GLfloat)height,0.1f,100.0f);	// Calculate The Aspect Ratio Of The Window
	gluLookAt(
		0.0, 0.0, 0.5,		// position
		0.0, 0.0, 0.0,		// looking at
		0.0, 1.0, 0.0			// up
	);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Rendering should fill screen/window completely, regardless of dimensions
	glScalef((width > height) ? (float)width / (float)height : 1.0, (height > width) ? (float)height / (float)width : 1.0, 1.0);
}

void render_grid(int horizontal, int vertical)
{
	int i;

	glColor4f(0.2, 0.2, 0.2, 1.0);
	glPushMatrix();
		glTranslatef(-0.5f, -0.5f, 0.0f);		// Put 0.0 in bottom left corner
		glEnable(GL_LINE_STIPPLE);
		glLineStipple(2, 0x0F0F);

		glBegin(GL_LINES);
		for(i=0 ; i<horizontal ; i++) {
			glVertex3f(0.0, (float)i / (float)horizontal, 0.0);
			glVertex3f(1.0, (float)i / (float)horizontal, 0.0);
		}
		for(i=0 ; i<vertical ; i++) {
			glVertex3f((float)i / (float)vertical, 0.0, 0.0);
			glVertex3f((float)i / (float)vertical, 1.0, 0.0);
		}
		glEnd();
		glDisable(GL_LINE_STIPPLE);
	glPopMatrix();
}

void render_samples(float* bar_heights, int count)
{
	int i;

	glPushMatrix();
		glTranslatef(-0.5f, -0.5f, 0.0f);		// Put 0.0 in bottom left corner

		glBegin(GL_QUADS);
		for(i=0 ; i<count ; i++) {
			float x1 = (GLfloat)i/(GLfloat)count;
			float y1 = bar_heights[i];

			float x2 = (GLfloat)(i+1)/(GLfloat)count;
			float y2 = 0.0;

			glColor4f(1.0, 0.8, 0.0, 0.4);
			glVertex3f(x2, y1, 0.0);
			glVertex3f(x2, y2, 0.0);

			glColor4f(0.9, 0.7, 0.0, 1.0);
			glVertex3f(x1, y2, 0.0);
			glVertex3f(x1, y1, 0.0);
		}
		glEnd();
	glPopMatrix();
}

//
// main()
//
int main(int argc, char* argv[])
{
	int width = 400; //800;
	int height = 400; //600;
	int sdl_flags = SDL_OPENGL | SDL_RESIZABLE;
	bool fullscreen = false;

	int next_arg = 1;
	if(argc >= 2 && strcmp(argv[next_arg], "--fullscreen") == 0) {
		fullscreen = true;

		// TODO: these numbers don't belong here
		width = 1024;
		height = 768;
		sdl_flags |= SDL_FULLSCREEN;
		next_arg++;
	}
	char* alsa_device_name = argv[next_arg];

	lo_address t = lo_address_new(UDP_ADDRESS, UDP_PORT);

	// TODO: this should be dynamic
	vector<Filter*> filters;
	filters.push_back(new Filter(-0.25, 0.45));
	filters.push_back(new Filter( 0.0,  0.45));
	filters.push_back(new Filter( 0.25, 0.45));

	int i;

	AudioSampler* pAudioSampler = new AudioSampler();
	pAudioSampler->Open(alsa_device_name);

	// Initialize SDL
	printf("Initializing SDL...\n");
	if(SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Unable to initialize SDL: %s\n", SDL_GetError());
		exit(1);
	}
	SDL_Surface* sdl_surface = SDL_SetVideoMode(width, height, 0, sdl_flags);

	if(sdl_surface == NULL) {
		fprintf(stderr, "Unable to create OpenGL screen: %s\n", SDL_GetError());
		SDL_Quit();
		exit(1);
	}

	SDL_WM_SetCaption(APPLICATION_NAME, NULL);
	SDL_WM_SetIcon(SDL_LoadBMP("icon.bmp"), NULL);
	initialize_OpenGL(width, height);

	int frames_per_second = 72;		// TODO: this doesn't belong here
	int ms_per_frame = 1000 / frames_per_second;

	SDL_Event event;
	bool done = false;
	bool pause_animation = false;
	bool pause = false;

	Filter* filter;

	while(!done) {
		int frame_time = SDL_GetTicks();

		// Check for input
		while(SDL_PollEvent(&event)) {
			switch(event.type) {
			case SDL_QUIT:
				{
					done = true;
				}
				break;

			case SDL_VIDEORESIZE:
				{
					width=event.resize.w;
					height=event.resize.h;

					SDL_FreeSurface(sdl_surface);
					sdl_surface = SDL_SetVideoMode(width, height, 0, sdl_flags);
					initialize_OpenGL(width, height);
				}
				break;

			case SDL_KEYDOWN:
				{
					switch(event.key.keysym.sym) {
					case SDLK_ESCAPE:
						if(fullscreen) {
							SDL_WM_ToggleFullScreen(sdl_surface);
							fullscreen = false;
						} else {
							SDL_WM_IconifyWindow();
						}
						break;
					case SDLK_F12:
						if(!fullscreen) {
							SDL_WM_ToggleFullScreen(sdl_surface);
							fullscreen = true;
						}
						break;
					case SDLK_SPACE:
						pause = !pause;
						break;
					}
				}
				break;

			case SDL_MOUSEBUTTONDOWN:
				// Send mouse presses to filters, until one returns 'true' to claim it
				{
					vector<Filter*>::iterator itFilter = filters.begin();
					for(; itFilter < filters.end(); itFilter++) {
						filter = *itFilter;
						if(filter->PointerPress(event.button.button, ((float)event.motion.x / (float)width) - 0.5, -(((float)event.motion.y / (float)height) - 0.5)))
							break;
					}
				}
				break;

			case SDL_MOUSEMOTION:
				{
					vector<Filter*>::iterator itFilter = filters.begin();
					for(; itFilter < filters.end(); itFilter++) {
						filter = *itFilter;
						filter->PointerMovement(((float)event.motion.x / (float)width) - 0.5, -(((float)event.motion.y / (float)height) - 0.5), ((float)event.motion.xrel / (float)width), -(((float)event.motion.yrel / (float)height)));
					}
				}
				break;

			case SDL_MOUSEBUTTONUP:
				{
					vector<Filter*>::iterator itFilter = filters.begin();
					for(; itFilter < filters.end(); itFilter++) {
						filter = *itFilter;
						filter->PointerRelease(event.button.button, ((float)event.motion.x / (float)width) - 0.5, -(((float)event.motion.y / (float)height) - 0.5));
					}
				}
				break;

			case SDL_ACTIVEEVENT:		// window focus changed
				{
					if(event.active.state & SDL_APPACTIVE) {
						pause_animation = (event.active.gain == 0);		// Pause animation when minimizing (continue analysis and sending OSC)
					}
				}
			}
		}

		// Get audio (NOTE: we can't skip this step or audio capture gets flaky
		pAudioSampler->Update();

		static int last_frame_time = 0;		// in milliseconds
		int frame_delta = frame_time - last_frame_time;

		// Time for a new frame?
		if(pause == false && frame_delta >= ms_per_frame) {
			//
			// Spectrum Analysis
			//
			pAudioSampler->Analyze();

			//
			// Update statuses of filters (the graphical boxes) and send OSC
			//
			{
				vector<Filter*>::iterator itFilter = filters.begin();
				for(i=0 ; itFilter < filters.end(); itFilter++, i++) {
					filter = *itFilter;

					// Update returns 'true' if activation value changed
					if(filter->Update(pAudioSampler->GetMagnitudeArray(), pAudioSampler->GetMagnitudeCount())) {
						snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "Spectrum Box %02d / Activation", i + 1);
						lo_send(t, address_buffer, "f", filter->GetActivation());
					}
				}
			}

			//
			// Render
			//
			if(pause_animation == false) {
				glClear(GL_COLOR_BUFFER_BIT);

				render_samples(pAudioSampler->GetMagnitudeArray(), pAudioSampler->GetMagnitudeCount());
				render_grid(NUM_HORIZONTAL_LINES, NUM_VERTICAL_LINES);

				// Render boxes
				vector<Filter*>::iterator itFilter = filters.begin();
				for(; itFilter < filters.end(); itFilter++) {
					filter = *itFilter;
					filter->Render();
				}

				SDL_GL_SwapBuffers();
			}

			//
			// End Frame
			//
			last_frame_time = frame_time;
		}
	}
	SDL_Quit();
	return 1;
}
