/*
 *   Copyright 2008 Ian McIntosh <ian@openanswers.org>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//
// Xinput Support for Tablets
//
#include <X11/Xlib.h>
#include <X11/extensions/XInput.h>
#include <X11/Xutil.h>

//
// liblo OpenSoundControl (http://liblo.sourceforge.net/)
//
#include <lo/lo.h>

#define APPLICATION_NAME ("Tablet to OSC")
#define APPLICATION_VERSION ("0.1")
#define APPLICATION_COPYRIGHT ("Copyright 2008 Ian McIntosh")

#define UDP_PORT ("10007") // LOOOZ :D
#define UDP_ADDRESS NULL //("255.255.255.255")

typedef int TBoolean;
#define TRUE (1)
#define FALSE (0)

typedef struct
{
	int min;
	int max;
} TLimits;

float scale_and_expand_limits(int value, TLimits* limits)
{
	if(value < limits->min)
		limits->min = value;

	if(value > limits->max)
		limits->max = value;

	int range = (limits->max - limits->min);

	if(range == 0)
		return 0.0;

	return (float)(value - limits->min) / (float)range;
}

#define MAX_DEVICE_AXES (6)
char* axes_names[MAX_DEVICE_AXES] = {
	"X",
	"Y",
	"Pressure",
	"Tilt X",
	"Tilt Y",
	""	// TODO: what is this?
};

TBoolean invert_axis[MAX_DEVICE_AXES] = {FALSE, TRUE, FALSE, FALSE, TRUE, FALSE};

#define INVALID_EVENT_TYPE (-1)

static int motion_type = INVALID_EVENT_TYPE;
static int button_press_type = INVALID_EVENT_TYPE;
static int button_release_type = INVALID_EVENT_TYPE;
static int proximity_in_type = INVALID_EVENT_TYPE;
static int proximity_out_type = INVALID_EVENT_TYPE;

static int register_events(Display *display, XDevice *device)
{
	int number = 0;		// number of events registered
	XEventClass event_list[7];

	if (device->num_classes > 0) {
		int i;
		for (i=0; i<device->num_classes; i++) {
			XInputClassInfo *ip = &(device->classes[i]);

			if(ip->input_class == ButtonClass) {
				DeviceButtonPress(device, button_press_type, event_list[number]); number++;
				DeviceButtonRelease(device, button_release_type, event_list[number]); number++;
			}
			else if (ip->input_class == ValuatorClass) {
				DeviceMotionNotify(device, motion_type, event_list[number]); number++;
				ProximityIn(device, proximity_in_type, event_list[number]); number++;
				ProximityOut(device, proximity_out_type, event_list[number]); number++;
			}
		}

		Window root_window = RootWindow(display, DefaultScreen(display));
		if (XSelectExtensionEvent(display, root_window, event_list, number)) {
			fprintf(stderr, "error selecting extended events\n");
			return 0;
		}
	}
	return number;
}

XDeviceInfo	* find_device_info(Display *display, char* device_name)
{
	int num_devices = 0;
	XDeviceInfo	*devices = XListInputDevices(display, &num_devices);

	int i;
	for(i=0; i<num_devices; i++) {
		if(strcmp(devices[i].name, device_name) == 0) {
			return &devices[i];
		}
	}
	return NULL;
}

#define DEFAULT_DEVICE_NAME ("stylus")

int main(int argc, char * argv[])
{
	printf("%s v%s %s\n", APPLICATION_NAME, APPLICATION_VERSION, APPLICATION_COPYRIGHT);

	char* device_name = argv[1];

	int device_number = 1;
	if(argc == 3) {
		device_number = atoi(argv[2]);
		printf("Sending as device %d...\n", device_number);
	}

	// 1) Connect to X server
	Display *display = XOpenDisplay(NULL);
	if(display == NULL) {
		fprintf(stderr, "unable to connect to X server\n");
		exit(1);
	}

	// 2) Check for presence of XInput extension
	XExtensionVersion *version = XGetExtensionVersion(display, INAME);
	if(version == NULL || version == (XExtensionVersion*) NoSuchExtension || version->present == False) {
		fprintf(stderr, "required extension '%s' missing in X server\n", INAME);
		exit(1);
	}
	XFree(version);

	// 3) Find device
	XDeviceInfo *device_info = find_device_info(display, device_name);
	if(device_info == NULL) {
		fprintf(stderr, "unable to find device '%s'\n", device_name);
		exit(1);
	}

	// 4) Open device
	XDevice *device = XOpenDevice(display, device_info->id);
	if(device == NULL) {
		fprintf(stderr, "unable to open device '%s'\n", device_info->name);
		exit(1);
	}

	// 5) Register events for device
	if(register_events(display, device) == 0) {
		fprintf(stderr, "unable to register events for device '%s'\n", device_info->name);
		exit(1);
	}

	// 6) OpenSoundControl address for sending messages
	lo_address t = lo_address_new(UDP_ADDRESS, UDP_PORT);

	#define ADDRESS_BUFFER_SIZE 1000
	char address_buffer[ADDRESS_BUFFER_SIZE + 1];

	int *axis_values = calloc(MAX_DEVICE_AXES, sizeof(int));
	TLimits *axis_limits = calloc(MAX_DEVICE_AXES, sizeof(TLimits));		// 0 for min and max is fine for joystick axes

	printf("Successfully connected to '%s'\n", device_info->name);
	printf("Ctrl-C to quit...\n");

	// 7) Loop and read events
	XEvent event;
	while(1) {
		XNextEvent(display, &event);

		if (event.type == motion_type) {
			XDeviceMotionEvent* motion_event = (XDeviceMotionEvent*)&event;

			// 7.1) Check each axis for change
			int axis_index;
			for(axis_index=0 ; axis_index < motion_event->axes_count && axis_index < MAX_DEVICE_AXES ; axis_index++) {
				int axis_value = motion_event->axis_data[axis_index];

				// 7.1.1) Send if axis changed
				if(axis_value != axis_values[axis_index]) {
					float value = scale_and_expand_limits(axis_value, &(axis_limits[axis_index]));

					// 7.1.1.1) Invert some axes so it looks like the cartesian plane
					if(invert_axis[axis_index])
						value = 1.0 - value;

					// 7.1.3) Format and pass to liblo (OpenSoundControl library)
					snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "Tablet %02d / %s", device_number, axes_names[axis_index]);
					lo_send(t, address_buffer, "f", value);

					axis_values[axis_index] = axis_value;		// NOTE: this saves the raw integer axis_value
				}
			}
		} else if(event.type == button_press_type) {
			XDeviceButtonEvent *button = (XDeviceButtonEvent *) &event;
			snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "Tablet %02d / Button %02d", device_number, button->button);
			lo_send(t, address_buffer, "i", 1);
		} else if(event.type == button_release_type) {
			XDeviceButtonEvent *button = (XDeviceButtonEvent *) &event;
			snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "Tablet %02d / Button %02d", device_number, button->button);
			lo_send(t, address_buffer, "i", 0);
		} else if(event.type == proximity_in_type) {
			snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "Tablet %02d / Pen Present", device_number);
			lo_send(t, address_buffer, "i", 1);
		} else if(event.type == proximity_out_type) {
			snprintf(address_buffer, ADDRESS_BUFFER_SIZE, "Tablet %02d / Pen Present", device_number);
			lo_send(t, address_buffer, "i", 0);
		}
	}
	XFlush(display);
}
