/*
 *   Copyright 2008 Ian McIntosh <ian@openanswers.org>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <stddef.h>
#include <math.h>

//
// Touchpad Support
//
#include <X11/Xdefs.h>
#include "synaptics.h"

//
// liblo OpenSoundControl (http://liblo.sourceforge.net/)
//
#include <lo/lo.h>

#define APPLICATION_NAME ("Synaptic Touchpad to OSC")
#define APPLICATION_VERSION ("0.2")
#define APPLICATION_COPYRIGHT ("Copyright 2008 Ian McIntosh")

#define UDP_PORT ("10007") // LOOOZ :D
#define UDP_ADDRESS (NULL) //("255.255.255.255")

typedef struct
{
	int min;
	int max;
} TLimits;

float scale_and_expand_limits(int value, TLimits* limits)
{
	if(value < limits->min)
		limits->min = value;

	if(value > limits->max)
		limits->max = value;

	int range = (limits->max - limits->min);

	if(range == 0)
		return 0.0;

	return (float)(value - limits->min) / (float)range;
}

int
main(int argc, char *argv[])
{
	printf("%s v%s %s\n", APPLICATION_NAME, APPLICATION_VERSION, APPLICATION_COPYRIGHT);

	SynapticsSHM *synshm;
	int shmid;

	/* Connect to the shared memory area */
	if ((shmid = shmget(SHM_SYNAPTICS, sizeof(SynapticsSHM), 0)) == -1) {
		if ((shmid = shmget(SHM_SYNAPTICS, 0, 0)) == -1) {
			fprintf(stderr, "Can't access shared memory area. SHMConfig disabled?\n");
			exit(1);
		} else {
			fprintf(stderr, "Incorrect size of shared memory area. Incompatible driver version?\n");
			exit(1);
		}
	}
	if ((synshm = (SynapticsSHM*) shmat(shmid, NULL, 0)) == NULL) {
		perror("shmat");
		exit(1);
	}

	printf("Connected to touchpad...\n");

	SynapticsSHM old = {0};

	// Where to send OpenSoundControl messages
	// If liblo < 0.25, send to NULL to at least get localhost
	// liblo >= 0.25 supports multicast at 255.255.255.255
	lo_address t = lo_address_new(UDP_ADDRESS, UDP_PORT);

	TLimits x_limits = {10000,0}, y_limits = {10000,0}, z_limits = {10000,0};

	printf("Ctrl-C to quit...\n");
	while (1) {
		SynapticsSHM cur = *synshm;

		// printf("x=%d,y=%d,z=%d\n", cur.x, cur.y, cur.z);

		// For some reason, when lifting a finger off a Synaptics Touchpad, the x sometimes jumps to 1 for the last few updates
		if (cur.x != 1 && cur.z >= 5) {
			if (cur.x != old.x) lo_send(t, "TouchPad / X", "f", scale_and_expand_limits(cur.x, &x_limits));

			// NOTE: y is inverted because ALPS and Synaptic both place y==0 at the top and we prefer a cartesian coordinate system
			if (cur.y != old.y) lo_send(t, "TouchPad / Y", "f", 1.0 - scale_and_expand_limits(cur.y, &y_limits));
		}

		// Z is a bit different
		if (cur.z != old.z) {
			// ALPS send either 0 or ~40->120+ for z.
			// We special-case 0 so scale_and_expand_limits never sees it, which prevents a leap to ~40/MAX when lightly touched.
			if (cur.z == 0) {
				lo_send(t, "TouchPad / Pressure", "f", 0.0);
				lo_send(t, "TouchPad / Touch", "i", 0);
			}
			else {
				float pressure = scale_and_expand_limits(cur.z, &z_limits);

				// Because of the special case for z==0 above, the minimum non-0 z value will be returned as 0.0.
				if(pressure == 0.0) { pressure = 0.008; }	// NOTE: this value is arbitrary (but approx 1/128 which is probably the max)

				lo_send(t, "TouchPad / Pressure", "f", pressure);
				lo_send(t, "TouchPad / Touch", "i", 1);
			}
		}

		// Copy current touchpad settings to old ones
		old = cur;

		usleep(5 * 1000);
	}
}
