 ###############################################################################
 #  Copyright 2006 Ian McIntosh <ian@openanswers.org>
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
 #  the Free Software Foundation; either version 2 of the License, or
 #  (at your option) any later version.
 #
 #  This program is distributed in the hope that it will be useful,
 #  but WITHOUT ANY WARRANTY; without even the implied warranty of
 #  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 #  GNU Library General Public License for more details.
 #
 #  You should have received a copy of the GNU General Public License
 #  along with this program; if not, write to the Free Software
 #  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 ###############################################################################

# Overwrite the method at /usr/lib/ruby/1.8/glib2.rb:37 (on Ubuntu Feisty), which
# has broken our ability to receive exceptions in our own code, and instead hands
# us a SystemExit exception (from the call to exit() on line 45) every time.
module GLib
	def exit_application(status, other)
		puts "(ignoring call to GLib::exit_application(#{status}), see #{__FILE__}:#{__LINE__} for details)"
		raise		# re-raise to be caught by our main loop exception handling below
	end
	module_function :exit_application
end

class GLib::Object
	alias :old_signal_connect :signal_connect

	def signal_connect(signal_name, &proc)
		s = old_signal_connect(signal_name) { |*args|
			begin
				proc.call(*args)
			rescue Exception => e
				e.report("in signal handler ('#{signal_name}')")
			end
		}
	end
end

module Gtk
	def self.main_clear_queue
		while events_pending? do main_iteration_do(blocking = false) end
	end

	def self.non_expanding(widget)
		Gtk::HBox.new.add_tight(widget).add(Gtk::Label.new(''))
	end

	def self.hbox_for_widgets(widgets, spacing=6, border=2)
		hbox = Gtk::HBox.new(homogeneous = false, spacing).set_border_width(border)
		widgets.each { |w| hbox.add_tight(w) }
		return hbox
	end

	def self.vbox_for_widgets(widgets, spacing=6, border=2)
		vbox = Gtk::VBox.new(homogeneous = false, spacing).set_border_width(border)
		widgets.each { |w| vbox.add_tight(w) }
		return vbox
	end

	def self.idle_add_once
		idle_add { yield ; false } # false = don't call again
	end
end

class Gtk::Container
	def add_tight(child)
		add(child)
		set_child_packing(child, expand=false, fill=false, padding=0, Gtk::PACK_START)
	end
end

class Gtk::Entry
	def on_change
		signal_connect('changed') { yield }
	end
end

class Gtk::Range
	def on_change
		signal_connect('value-changed') { yield }
	end
end

class Gtk::TreeSelection
	def selected_iters
		a = []
		selected_each { |model, path, iter| a << iter }
		return a
	end

	def empty?
		selected_each { return false }
		return true
	end

	def each_iter
		selected_each { |model, path, iter| yield iter }
	end
end

module Gtk::TreeSortable
	def sorted?
		sort_column_id != nil
	end
end

class Gtk::TreeView
	DRAG_ROW_DISTANCE = 15

	def initialize
		super
		signal_connect('key-press-event') { |obj, event|
			# Ctrl-Up and Ctrl-Down move selected rows
			if !model.sorted?
				if event.keyval == Gdk::Keyval::GDK_Up and event.state == Gdk::Window::ModifierType::CONTROL_MASK
					move_selected_up
					true # handled
				elsif event.keyval == Gdk::Keyval::GDK_Down and event.state == Gdk::Window::ModifierType::CONTROL_MASK
					move_selected_down
					true # handled
				end
			end
			false
		}

		# Basic drag-and-drop
		on_primary_mouse_button_down { |x, y|
			unless (selection.empty? or model.sorted?)
				@tree_view_rows_grabbed, @tree_view_grab_y = true, y
				Gtk.grab_add(self)
			end
		}
		on_mouse_motion { |x, y|
			if @tree_view_rows_grabbed
				offset = (y - @tree_view_grab_y)
				if offset > DRAG_ROW_DISTANCE		# TODO: this should be a row measurement, not hardcoded
					@tree_view_grab_y = y if move_selected_down	# reset y if we moved something (so dragging the last row down a lot then up a little doesn't immediately move the row up)
				elsif offset < -DRAG_ROW_DISTANCE
					@tree_view_grab_y = y if move_selected_up
				end
			end
		}
		on_primary_mouse_button_up {
			@tree_view_rows_grabbed = false
			Gtk.grab_remove(self)
		}
	end

	def cancel_row_grab
		@tree_view_rows_grabbed = false
	end

	def on_order_change(&proc)
		@on_order_change_proc = proc
	end

	def move_selected_up
		sleep 0.01	# this seems to prevent a segfault when dragging rows

		# if can't do 'prev!' then the first row is at the top, so we do nothing
		selection.selected_iters.each { |iter| path = iter.path ; return false unless path.prev! ; prev_iter = model.get_iter(path) ; model.swap(iter, prev_iter) }
		@on_order_change_proc.call if @on_order_change_proc
		return true
	end

	def move_selected_down
		sleep 0.01	# this seems to prevent a segfault when dragging rows

		# if can't do 'next!' then the last row (processed first) is at the bottom, so we do nothing
		selection.selected_iters.reverse.each { |iter| next_iter = iter.dup ; return false unless next_iter.next! ; model.swap(next_iter, iter) }
		@on_order_change_proc.call if @on_order_change_proc
		return true
	end

	def clear
		model.clear
	end

	def on_selection_change
		selection.signal_connect('changed') { yield }
	end

	def each
		model.each { |model, path, iter| yield model, path, iter }
	end

	def each_iter
		model.each {|model, path, iter| yield iter }
	end

	def unselect_all
		selection.unselect_all
		self
	end

	def on_activated
		signal_connect('row-activated') { yield }
	end
end

class Gtk::ListStore
	def each_iter
		each { |model, path, iter| yield iter }
	end

	def empty?
		each { return false }	# TODO: better way to do this?
		return true
	end
end

class Gtk::CellRendererToggle
	def on_toggled
		signal_connect('toggled') { |renderer, path| yield path }
	end
end

class Gtk::Table
	def add_row
		resize(n_rows + 1, n_columns)
		return n_rows - 1
	end
end

class Gtk::ColorSelection
	def gl_color
		return [current_color.red, current_color.green, current_color.blue, current_alpha].collect { |n| n.to_f / 65535.0 } if has_opacity_control?
		return [current_color.red, current_color.green, current_color.blue].collect { |n| n.to_f / 65535.0 }
	end
end

class Gtk::ColorButton
	def gl_color
		return [color.red, color.green, color.blue, alpha].collect { |n| n.to_f / 65535.0 } if use_alpha?
		return [color.red, color.green, color.blue].collect { |n| n.to_f / 65535.0 }
	end

	def gl_color=(c)
		set_color(Gdk::Color.new(c[0] * 65535.0, c[1] * 65535.0, c[2] * 65535.0))
		set_alpha(c[3] * 65535.0) if use_alpha? and c[3]
	end

	def gl_alpha
		return alpha.to_f / 65535.0
	end

	def on_change
		signal_connect('color-set') { yield }
	end
end

class Gtk::ToggleButton
	def on_change_with_init
		yield		# yield once now to init
		on_change { yield }
	end

	def toggle
		set_active(!active?)
	end

	def on_change
		signal_connect('toggled') { yield }
	end
end

class Gtk::Window
	def toggle_fullscreen
		@is_fullscreen = !@is_fullscreen
		fullscreen if @is_fullscreen
		unfullscreen if not @is_fullscreen
	end

	def toggle_visibility
		if visible?
			hide
		else
			present
		end
	end

	def on_lose_focus
		signal_connect('focus-out-event') { yield }
	end

	def present_modal
		self.modal = true
		present
	end

	def choose_file_save_path(title = 'Save File', filepath = nil, pattern = nil)
		dialog = Gtk::FileChooserDialog.new(title, self, Gtk::FileChooser::ACTION_SAVE, nil, [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL], [Gtk::Stock::SAVE, Gtk::Dialog::RESPONSE_ACCEPT])
		dialog.set_filter(Gtk::FileFilter.new.add_pattern(pattern)) if pattern
		dialog.set_filename(filepath) if filepath

		dialog.set_do_overwrite_confirmation(true)
		if dialog.run == Gtk::Dialog::RESPONSE_ACCEPT
			filename = dialog.filename
			dialog.hide
			yield filename
		end
		dialog.destroy
	end

	def choose_file_open_path(title = 'Open File', directory = nil, pattern = nil)
		dialog = Gtk::FileChooserDialog.new(title, self, Gtk::FileChooser::ACTION_OPEN, nil, [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL], [Gtk::Stock::OPEN, Gtk::Dialog::RESPONSE_ACCEPT])
		dialog.set_filter(Gtk::FileFilter.new.add_pattern(pattern)) if pattern
		dialog.set_current_folder(directory) if directory

		if dialog.run == Gtk::Dialog::RESPONSE_ACCEPT
			filename = dialog.filename
			dialog.hide
			yield filename
		end
		dialog.destroy
	end
end

class Gtk::Widget
	def on_expose
		signal_connect('expose-event') { yield ; true }		# false = unhandled
	end

	def on_focus
		signal_connect('focus-in-event') { yield ; true }		# false = unhandled
	end

	def toggle_visibility
		self.set_visible(!self.visible?)
	end

	# Button presses / releases
	def on_button_press
		add_events(Gdk::Event::BUTTON_PRESS_MASK)
		signal_connect('button-press-event') { |obj, event| yield event if event.event_type == Gdk::Event::BUTTON_PRESS ; false }		# false = unhandled
	end

	def on_button_release
		add_events(Gdk::Event::BUTTON_RELEASE_MASK)
		signal_connect('button-release-event') { |obj, event| yield event if event.event_type == Gdk::Event::BUTTON_RELEASE ; false }		# false = unhandled
	end

	def on_double_click
		add_events(Gdk::Event::BUTTON_PRESS_MASK)
		signal_connect('button-press-event') { |obj, event| yield event if event.event_type == Gdk::Event::BUTTON2_PRESS ; false }		# false = unhandled
	end

	def on_scroll_wheel
		add_events(Gdk::Event::SCROLL_MASK)
		signal_connect('scroll-event') { |obj, event| yield event ; false }		# false = unhandled
	end

	def on_scroll_wheel_up
		add_events(Gdk::Event::SCROLL_MASK)
		on_scroll_wheel { |event| yield if event.direction == Gdk::EventScroll::UP ; false }		# false = unhandled
	end

	def on_scroll_wheel_down
		add_events(Gdk::Event::SCROLL_MASK)
		on_scroll_wheel { |event| yield if event.direction == Gdk::EventScroll::DOWN ; false }		# false = unhandled
	end

	def on_primary_mouse_button_down
		on_button_press { |event| yield event.x, event.y if (event.event_type == Gdk::Event::BUTTON_PRESS and event.button == 1) ; false }
	end
	alias :on_click :on_primary_mouse_button_down

	def on_primary_mouse_button_up
		on_button_release { |event| yield event.x, event.y if (event.event_type == Gdk::Event::BUTTON_RELEASE and event.button == 1) }
	end

	def on_primary_mouse_button_double_click
		on_button_press { |event| yield event.x, event.y if (event.event_type == Gdk::Event::BUTTON_PRESS and event.button == 1) }
	end
	alias :on_click :on_primary_mouse_button_down

	def on_secondary_mouse_button_down
		on_button_press { |event| yield event.x, event.y if (event.event_type == Gdk::Event::BUTTON_PRESS and event.button == 3) }
	end
	alias :on_click :on_primary_mouse_button_down

	def on_secondary_mouse_button_up
		on_button_release { |event| yield event.x, event.y if (event.event_type == Gdk::Event::BUTTON_RELEASE and event.button == 3) }
	end

	def on_context_menu
		on_button_press { |event|
			if (event.event_type == Gdk::Event::BUTTON_PRESS and event.button == 3)
				if (path = get_path_at_pos(event.x, event.y))
					if selected.size <= 1
						unselect_all
						selection.select_path(path[0])
					end
				end
				yield event
			end
		}
	end

	def on_mouse_motion
		add_events(Gdk::Event::POINTER_MOTION_MASK)
		signal_connect('motion-notify-event') { |obj, event| yield event.x, event.y ; false } # false = unhandled
	end

	def width
		allocation.width
	end

	def height
		allocation.height
	end

	def queue_draw
		queue_draw_area(0, 0, allocation.width, allocation.height)
	end

	def on_key_press_event(event)
		modifiers = nil		# TODO: extract from event ?
		if @key_press_handlers and @key_press_handlers[modifiers] and @key_press_handlers[modifiers][event.keyval]
			@key_press_handlers[modifiers][event.keyval].each { |proc| proc.call(event) }
			true
		else
			false
		end
	end

	def on_key_press(keyval, modifiers = nil, &proc)
		if @key_press_handlers.nil?
			# On first use of on_key_press() setup GLib signal handler
			self.signal_connect('key-press-event') { |obj, event| on_key_press_event(event) }
			@key_press_handlers = {}
		end

		@key_press_handlers[modifiers] ||= {}
		@key_press_handlers[modifiers][keyval] ||= []
		@key_press_handlers[modifiers][keyval] << proc
	end

	def on_activate
		signal_connect('activate') { yield }
	end

	def on_close
		signal_connect('delete_event') { yield ; true }		# true = handled
	end
end

class Gtk::ComboBox
	def on_change_with_init
		yield		# yield once now to init
		on_change { yield }
	end

	def each_iter
		model.each {|model, path, iter| yield iter }
	end

	def select_first
		self.active_iter = model.iter_first
	end

	def on_change
		signal_connect('changed') { yield }
	end
end

class Gtk::TextView
	def text=(rhs)
		buffer.text = rhs
	end
end

class Gtk::Notebook
	def on_change
		signal_connect('switch-page') { |widget, page, page_number| yield page_number }
	end

	def on_change_with_init
		yield self.page		# yield once now to init
		on_change { |page_number| yield page_number }
	end
end

class Gtk::SpinButton
	def on_change
		signal_connect('value-changed') { yield }
	end
end

class Gtk::Button
	def on_click
		signal_connect('clicked') { yield }
	end
end

class Gtk::Viewport
	def draggable!
		on_primary_mouse_button_down { |x, y| @grabbed, @grab_x, @grab_y = true, x, y ; window.set_cursor(Gdk::Cursor.new(Gdk::Cursor::FLEUR))  }
		on_mouse_motion { |x, y|
			if @grabbed
				delta = (x - @grab_x)
				max_scroll = hadjustment.upper - allocation.width		# total size - viewport size
				hadjustment.value = (hadjustment.value - delta).clamp(0, max_scroll + 2)

				delta = (y - @grab_y)
				max_scroll = vadjustment.upper - allocation.height		# total size - viewport size
				vadjustment.value = (vadjustment.value - delta).clamp(0, max_scroll + 2)
			end
		}
		on_primary_mouse_button_up { @grabbed = false ; window.set_cursor(Gdk::Cursor.new(Gdk::Cursor::ARROW)) }
	end
end

class String
	def with_pango_tag(tag)
		return "<#{tag}>#{self}</#{tag}>"
	end

	def with_optional_pango_tag(bool, tag)
		return self.with_pango_tag(tag) if bool
		return self
	end
end
