 ###############################################################################
 #  Copyright 2008 Ian McIntosh <ian@openanswers.org>
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
 #  the Free Software Foundation; either version 2 of the License, or
 #  (at your option) any later version.
 #
 #  This program is distributed in the hope that it will be useful,
 #  but WITHOUT ANY WARRANTY; without even the implied warranty of
 #  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 #  GNU Library General Public License for more details.
 #
 #  You should have received a copy of the GNU General Public License
 #  along with this program; if not, write to the Free Software
 #  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 ###############################################################################

# Mixin: Drawing.
#
# IMPORTANT: This mixin MUST be included in your class BEFORE inheriting from your class.
# IMPORTANT: Any self.inherited method in your class MUST call 'super'.
#
# A typical and safe use is as follows:
#
#  class MyClass
#    include Drawing
#    ...
#    self.inherited(klass)
#      ...
#      super
#    end
#    ...
#  end

module Drawing
	def self.included(base) #:nodoc:
		base.extend(ClassMethods)
	end

	def self.conditional(*methods)
		methods.each { |method|
			self.class_eval <<-end_class_eval
				def #{method}_if(bool, *args, &proc)
					if bool
						#{method}(*args, &proc)
					else
						yield
					end
				end
			end_class_eval
		}
	end

	###################################################################
	#
	# Class-level methods
	#
	###################################################################
	module ClassMethods
	end

	###################################################################
	#
	# Object-level methods
	#
	###################################################################

	def clear_screen(color)
		GL.ClearColor(*(color.to_a))
		GL.Clear(GL::COLOR_BUFFER_BIT | GL::DEPTH_BUFFER_BIT)
	end

	def fade_screen_to_color(color, amount)		# This turned out to be a bit slower
		return if amount == 0.0
		return clear_screen(color) if amount == 1.0

		a = color.to_a
		with_color_and_alpha([a[0], a[1], a[2], 1.0]) {
			GL.DepthMask(GL::FALSE)
			with_identity_transformation {
				with_pixel_combine_function(:constant_alpha) {
					GL.BlendColor(0,0,0,amount)		# Set the "constant alpha"
					fullscreen_rectangle
					GL.BlendColor(0,0,0,0)
				}
			}
			GL.DepthMask(GL::TRUE)
		}
		GL.Clear(GL::DEPTH_BUFFER_BIT)
	end

	def fade_screen_to_color_with_alpha_blend(color, amount)
		return if amount == 0.0
		return clear_screen(color) if amount == 1.0

		a = color.to_a
		with_color_and_alpha([a[0], a[1], a[2], amount]) {
			GL.DepthMask(GL::FALSE)
			with_identity_transformation {
				fullscreen_rectangle
			}
			GL.DepthMask(GL::TRUE)
		}
		GL.Clear(GL::DEPTH_BUFFER_BIT)
	end

	def with_identity_transformation
		GL.PushMatrix		# push modelview
		GL.MatrixMode(GL::PROJECTION)
		GL.PushMatrix		# push projection

		$engine.projection		# TODO: shouldn't reference $engine here, instead have a way to store defaults
		$engine.view

		yield

		GL.MatrixMode(GL::PROJECTION)
		GL.PopMatrix		# pop projection
		GL.MatrixMode(GL::MODELVIEW)
		GL.PopMatrix
	end

	###################################################################
	# Moving and spinning around in 3D space
	###################################################################
	def translate(x, y, z=0.0)
		if x == 0.0 and y == 0.0 and z == 0.0
			yield
		else
			GL.SaveMatrix {
				GL.Translate(x, y, z)
				yield
				# Alternate non push/pop method: GL.Translate(-x, -y, -z)
			}
		end
	end
	conditional :translate

	# Variation of translate, moves "forward" (towards higher Y values)
	def slide(amount)
		if amount == 0.0
			yield
		else
			GL.SaveMatrix {
				GL.Translate(0.0, amount, 0.0)
				yield
				# Alternate non push/pop method: GL.Translate(0, -amount, 0)
			}
		end
	end
	conditional :slide

	def angle_slide(angle, distance)
		return yield if distance == 0.0

		# Rotate, send object out to its orbit, then rotate the other way so
		# the object doesn't appear to have rotated at all.
		GL.SaveMatrix {
			GL.Rotate(angle * FUZZY_TO_DEGREES, 0.0, 0.0, 1.0) 	# Rotate around the Z axis
			GL.Translate(0, distance, 0)
			GL.Rotate(-angle * FUZZY_TO_DEGREES, 0.0, 0.0, 1.0)	# Inverse of above
			yield
		}
	end
	conditional :angle_slide

	def rotate(amount, x=0.0, y=0.0, z=1.0)
		return yield if amount == 0.0
		GL.SaveMatrix {			# Because we can't reverse a scale to 0.0
			GL.Rotate(amount * FUZZY_TO_DEGREES, x, y, z) 	# Rotate around the Z axis
			yield
			# Alternate non push/pop method: GL.Rotate(-amount * FUZZY_TO_DEGREES, 0.0, 0.0, 1.0)
		}
	end
	alias :roll :rotate

	def pitch(amount)
		return yield if amount == 0.0

		GL.SaveMatrix {
			GL.Rotate(amount * FUZZY_TO_DEGREES, 1.0, 0.0, 0.0)
			yield
			# Alternate non push/pop method: GL.Rotate(-amount * FUZZY_TO_DEGREES, 1.0, 0.0, 0.0)
		}
	end
	alias :with_pitch :pitch

	def yaw(amount)
		return yield if amount == 0.0

		GL.SaveMatrix {
			GL.Rotate(amount * FUZZY_TO_DEGREES, 0.0, 1.0, 0.0)
			yield
			# Alternate non push/pop method: GL.Rotate(-amount * FUZZY_TO_DEGREES, 0.0, 1.0, 0.0)
		}
	end
	conditional :rotate, :roll, :pitch, :yaw
	alias :with_yaw :yaw

	def scale(x, y=nil)
		y ||= x							# When only one is given, scale equally x and y

		# Special-case very common values...
		return yield if x == 1.0 and y == 1.0

		GL.SaveMatrix {			# Because we can't reverse a scale to 0.0
			GL.Scale(x, y, 1.0)
			yield
		}
	end
	conditional :scale

	def distance_scale(amount, &proc)
		scale(1.0 - amount, &proc)
	end
	conditional :distance_scale

	def flip_vertically(&proc)
		scale(1.0, -1.0, &proc)
	end
	conditional :flip_vertically

	def flip_horizontally(&proc)
		scale(-1.0, 1.0, &proc)
	end
	conditional :flip_horizontally

	def unit_square
		@unit_square_list ||= GL.RenderToList {
			GL.Begin(GL::TRIANGLE_FAN)
				GL.TexCoord(0.0, 0.0) ; GL.Vertex(-0.5, 0.5)
				GL.TexCoord(1.0, 0.0) ; GL.Vertex(0.5, 0.5)
				GL.TexCoord(1.0, 1.0) ; GL.Vertex(0.5, -0.5)
				GL.TexCoord(0.0, 1.0) ; GL.Vertex(-0.5, -0.5)
			GL.End
		}
		GL.CallList(@unit_square_list)
	end
	alias :fullscreen_rectangle :unit_square

	def unit_square_outline
		@unit_square_outline_list ||= GL.RenderToList {
			GL.Begin(GL::LINE_LOOP)
				GL.Vertex(-0.5, 0.5)
				GL.Vertex(0.5, 0.5)
				GL.Vertex(0.5, -0.5)
				GL.Vertex(-0.5, -0.5)
			GL.End
		}
		GL.CallList(@unit_square_outline_list)
	end
	alias :fullscreen_rectangle :unit_square

	###################################################################
	# Clipping
	###################################################################

	def with_clip_plane(plane)
		$next_clip_plane_index ||= 0

		GL.Enable(GL::CLIP_PLANE0 + $next_clip_plane_index)
		GL.ClipPlane(GL::CLIP_PLANE0 + $next_clip_plane_index, plane)
		$next_clip_plane_index += 1
		yield
		$next_clip_plane_index -= 1
		GL.Disable(GL::CLIP_PLANE0 + $next_clip_plane_index)
	end
	conditional :with_clip_plane

	def with_clip_plane_left_of_angle(fuzzy_angle, &proc)
		radians = RADIANS_UP + (fuzzy_angle * RADIANS_PER_CIRCLE)
		with_clip_plane([-Math.sin(radians), -Math.cos(radians), 0.0, 0.0], &proc)
	end

	def with_clip_plane_right_of_angle(fuzzy_angle, &proc)
		radians = RADIANS_UP + (fuzzy_angle * RADIANS_PER_CIRCLE)
		with_clip_plane([Math.sin(radians), Math.cos(radians), 0.0, 0.0], &proc)
	end

	def with_vertical_clip_plane_left_of(x, &proc)
		with_clip_plane([1.0, 0.0, 0.0, -x], &proc)
	end

	def with_vertical_clip_plane_right_of(x, &proc)
		with_clip_plane([-1.0, 0.0, 0.0, x], &proc)
	end

	def with_horizontal_clip_plane_above(y, &proc)
		with_clip_plane([0.0, -1.0, 0.0, y], &proc)
	end

	def with_horizontal_clip_plane_below(y, &proc)
		with_clip_plane([0.0, 1.0, 0.0, -y], &proc)
	end

	###################################################################
	# Setting Attributes
	###################################################################

	def with_color(color)
		saved = GL.GetColorArray
		c = color.to_a
		GL.Color([c[0], c[1], c[2], (c[3] || 1.0) * (saved[3] || 1.0)])		# NOTE: doesn't set alpha-- instead multiplies it
		yield
		GL.Color(*saved)
	end
	conditional :with_color

	def with_color_and_alpha(color)
		saved = GL.GetColorArray
		GL.Color(*color.to_a)
		yield
		GL.Color(*saved)
	end
	conditional :with_color_and_alpha

	def with_alpha(alpha)
		saved = GL.GetColorArray
		#GL.Color(*color.to_a)
		GL.Color([saved[0], saved[1], saved[2], alpha])
		yield
		GL.Color(*saved)
	end

	def multiply_alpha(multiplier)
		return yield if multiplier == 1.0

		saved = GL.GetDoublev(GL::CURRENT_COLOR)
		c = saved.dup			# TODO: can we avoid this?
		c[3] *= multiplier
		GL.Color(*c)
		yield
		GL.Color(*saved)
	end

	DRAW_METHOD_OPTIONS = [[:average, 'Average'], [:brighten, 'Brighten'], [:darken, 'Darken']] #, [:test, 'Test']]
	def with_pixel_combine_function(name)
		case name
		#when :replace then GL.BlendFunc(GL::SRC_ALPHA, GL::ZERO)
		when :brighten then GL.BlendFunc(GL::SRC_ALPHA, GL::ONE)
		#when :brighten then GL.BlendFunc(GL::DST_COLOR, GL::SRC_ALPHA)
		#when :multiply then GL.BlendFunc(GL::DST_COLOR, GL::ZERO)
		when :darken then GL.BlendFunc(GL::DST_COLOR, GL::ZERO) #GL::SRC_ALPHA)
		#GL.BlendFunc(GL::ONE_MINUS_DST_COLOR, GL::ONE_MINUS_DST_COLOR)

		when :constant_alpha then GL.BlendFunc(GL::CONSTANT_ALPHA, GL::ONE_MINUS_CONSTANT_ALPHA)

		else # default
			yield
			return
		end
		yield
		GL.BlendFunc(GL::SRC_ALPHA, GL::ONE_MINUS_SRC_ALPHA)
	end

	###################################################################
	# Getting Attributes
	###################################################################

	def current_color_array
		GL.GetDoublev(GL::CURRENT_COLOR)
	end

	def current_color
		Color.new(current_color_array)
	end

	###################################################################
	# Texture Methods
	###################################################################

	def with_texture(id)
		return yield unless id

		GL.BindTexture(GL::TEXTURE_2D, id)
		yield
		GL.BindTexture(GL::TEXTURE_2D, 0)
	end

	def with_texture_scale(x_scale, y_scale)
		GL.MatrixMode(GL::TEXTURE)
			GL.SaveMatrix {
				GL.Scale(x_scale, y_scale, 1)

				GL.MatrixMode(GL::MODELVIEW)
					yield
				GL.MatrixMode(GL::TEXTURE)
			}
		GL.MatrixMode(GL::MODELVIEW)
	end

	def with_texture_scale_and_translate(x_scale, y_scale, x_translate, y_translate)
		GL.MatrixMode(GL::TEXTURE)
			GL.SaveMatrix {
				GL.Scale(x_scale, y_scale, 1)
				GL.Translate(x_translate, y_translate, 0)

				GL.MatrixMode(GL::MODELVIEW)
					yield
				GL.MatrixMode(GL::TEXTURE)
			}
		GL.MatrixMode(GL::MODELVIEW)
	end

	###################################################################
	# Framebuffer Objects
	###################################################################

	require 'framebuffer_object'

	def with_offscreen_buffer
		@fbo_list ||= []

		if @fbo_list.empty?
			fbo = FramebufferObject.new
		else
			fbo = @fbo_list.pop
		end

		yield fbo

		@fbo_list << fbo
	end

	###################################################################
	# Fuzzy Math (that is, between 0.0 and 1.0)
	###################################################################

	# Accepts 0.0 -> 1.0, returns 0.0 -> 1.0
	# (fuzzy 0.0 == radian 0.0)
	def fuzzy_sine(fuzzy)
		Math.sin(fuzzy * (Math::PI * 2.0)) / 2.0 + 0.5
	end

	def fuzzy_cosine(fuzzy)
		Math.cos(fuzzy * (Math::PI * 2.0)) / 2.0 + 0.5
	end

	def with_alpha_test(alpha_cutoff)
		GL.Enable(GL::ALPHA_TEST)
		GL.AlphaFunc(GL::GREATER, alpha_cutoff)
		yield
		GL.Disable(GL::ALPHA_TEST)
	end

	def with_stencil_buffer
		GL.Enable(GL::STENCIL_TEST)
		yield
		GL.Disable(GL::STENCIL_TEST)
	end

	#
	# Stencil operations
	#
	def with_stencil_buffer_for_writing(options = {})
		$next_stencil_buffer_index ||= 0

		# Select stencil buffer
		with_stencil_buffer {
			GL.StencilMask(1 << $next_stencil_buffer_index)

			# Clear stencil buffer
			GL.ClearStencil(0)
			GL.Clear(GL::STENCIL_BUFFER_BIT)

			GL.StencilOp(GL::REPLACE, GL::REPLACE, GL::REPLACE)
			GL.StencilFunc(GL::ALWAYS, 0x00000001, 0x00000001)

			# Don't write to color buffer
			GL.ColorMask(false, false, false, false)
			GL.DepthMask(false)

			with_alpha_test(options[:alpha_cutoff] || 0.0) {
				$next_stencil_buffer_index += 1
				yield
				$next_stencil_buffer_index -= 1
			}

			GL.DepthMask(true)
			GL.ColorMask(true, true, true, true)
			GL.StencilMask(0x00000000)
		}
	end

	def with_stencil_buffer_filter
		$next_stencil_buffer_index ||= 0

		with_stencil_buffer {
			GL.StencilMask(1 << $next_stencil_buffer_index)
			GL.StencilFunc(GL::EQUAL, 0x00000001, 0x00000001)

			$next_stencil_buffer_index += 1
			yield
			$next_stencil_buffer_index -= 1

			GL.StencilMask(0x00000000)
		}
	end

	#
	# Hit Testing
	#
	def with_hit_test
		$engine.with_env(:hit_test, true) {
			$hit_test_id = 0
			$hit_test_options = {}
			yield
		}
	end

	def next_hit_test_id
		$hit_test_id += 1
		return $hit_test_id
	end

	HANDLE_POSITION = 1
	def add_hit_test_option(hit_test_id, handle_id, object, x, y, z)
		$hit_test_options[[hit_test_id, handle_id]] = object
		GL.Color4ub(hit_test_id, HANDLE_POSITION, 0, 255)
		GL.Begin(GL::POINTS) ; GL.Vertex(x, y, z) ; GL.End
	end

	# returns [hit_test_id, handle_id] or [0, nil]
	def hit_test_object_handles(x, y)
		color = glReadPixels(x, y, 1, 1, GL_RGB, GL_UNSIGNED_BYTE).unpack("CCC")
		object = $hit_test_options[[color[0], color[1]]]
		return [object, color[1]]
	end
end
