require 'gl_widget'
require 'bsearch'

class GLCurveEditor < GLWidget
	Sample = Struct.new(:time, :value)

	attr_accessor :samples, :name

	def initialize(*args)
		super(*args)
		@samples = []
		@last_y = 0.0
	end

	def time_range_to_index_range(start_time, stop_time)
		@samples.bsearch_range { |s| (s.time < start_time) ? -1 : ((s.time > stop_time) ? 1 : 0) }
	end

	def draw_grid(x_count, y_count)
		@grid_display_lists ||= {}
		@grid_display_lists[x_count] ||= {}

		@grid_display_lists[x_count][y_count] ||= GL.RenderToList {
			GL.Begin(GL::LINES)
				(-0.5).step(0.5, 1.0 / x_count) { |x| GL.Vertex(x, 0.5) ; GL.Vertex(x, -0.5) }
				(-0.5).step(0.5, 1.0 / y_count) { |y| GL.Vertex(0.5, y) ; GL.Vertex(-0.5, y) }
			GL.End
		}
		GL.CallList(@grid_display_lists[x_count][y_count])
	end

	def draw(percentage, range)
		percent_begin = percentage - range/2.0
		percent_end = percent_begin + range

		index_range = time_range_to_index_range(percent_begin, percent_end)

		index_range = ((index_range.first-1)...(index_range.last)) if index_range.first > 0
		index_range = ((index_range.first)...(index_range.last+1)) if index_range.last < @samples.size-1

		current_index = time_to_sample_index(percentage)

		with_unit_scale {
			#
			# Background
			#
			with_color([@grab ? 0.2 : 0.1, @grab ? 0.5 : 0.4, @grab ? 0.2 : 0.1]) {
				unit_square
			}

			# respect zoom and progress
			scale(1.0 / range, 1.0) {
				translate(-percentage, 0.0) {

					#
					# Grid
					#
					with_color_and_alpha([0.0, 0.0, 0.0, 0.1]) {
						draw_grid(100, 8)
						translate(1.0, 0.0) { draw_grid(100, 8) }
					}

					#
					# Samples
					#
					translate(0.0, -0.5) {
						with_color([1.0,1.0,1.0]) {

#							vertices = [0.0,0.0, 0.5,0.2, 0.7,0.7, 1.0,1.0]
#							GL.VertexPointer(2, GL::FLOAT, 0, vertices)
#							GL.DrawArrays(GL::LINE_STRIP, 0, 4)

							GL.Begin(GL::LINE_STRIP)
								x = 0.0
								y = @samples[index_range.first].value rescue 0.0
								for i in index_range
									s = @samples[i]
									GL.Vertex(s.time, y)

									x = s.time
									y = s.value

									GL.Vertex(x, y)
								end

								if index_range.last == @samples.size and @samples.last and @samples.last.time < percentage
									# draw final point out to infinity, showing the user that the last sample value continues on
									GL.Vertex(percentage, y)
								end
							GL.End
						}
					}
				}
			}

			with_color_and_alpha([1.0,1.0,0.0,0.8]) {
				GL.Begin(GL::LINES)
					GL.Vertex(0.0, -0.5)
					GL.Vertex(0.0, 0.5)
				GL.End


				if current_index and @samples[current_index]
					current_value = @samples[current_index].value

					with_color_and_alpha([1.0,1.0,0.0,0.8]) {
						GL.PointSize(8.0)
						GL.Begin(GL::POINTS)
							GL.Vertex(0.0, current_value - 0.5)
						GL.End
					}
				end
			}
		}
	end

	def time_to_sample_index(time)
		i = @samples.bsearch_lower_boundary { |s| s.time <=> time }

		# exact match?
		return i if @samples[i] and @samples[i].time == time

		return i-1 # from boundary to index (see bsearch.png)
	end

	def add_sample(time, value)
		if @samples.empty?
			@samples << Sample.new(time, value)
		else
			i = time_to_sample_index(time)

			if @previous_update_time
				previous_i = time_to_sample_index(@previous_update_time)

				if previous_i < i
					# overwrite entries with a new value, NOT including previous_i
					@samples[(previous_i + 1), (i - previous_i)] = Sample.new(time, value) #, Sample.new(time, @samples[i].value)]

				# otherwise same sample. new value?
				elsif @samples[previous_i].value != value
					# insert after previous_i
					@samples.insert(previous_i+1, Sample.new(time, value))
				end

			elsif time == @samples[i].time
				# Update previous entry with latest value
				@samples[i].value = value

			else
				@samples.insert(i+1, Sample.new(time, value))
			end

			@previous_update_time = time
		end
	end

	def button_down(time, number, x, y)
		return unless within?(x,y)

		@grab = true
		x, y = global_to_local(x, y)
		@last_y = y
		add_sample(time, y)
	end

	def cursor_move(time, x, y)
		if @grab
			x, y = global_to_local(x, y)
			@last_y = y
			add_sample(time, y)
		end
	end

	def button_up(time, number, x, y)
		@grab = false
	end

	def update(time)
		if @grab
			add_sample(time, @last_y)
		else
			@previous_update_time = nil
		end
	end
end
