require 'gl_widget'
require 'bsearch'

class GLEventEditor < GLWidget
	Sample = Struct.new(:time, :value)

	attr_accessor :samples, :name

	def initialize(*args)
		super(*args)
		@samples = []
	end

	def time_range_to_index_range(start_time, stop_time)
		@samples.bsearch_range { |s| (s.time < start_time) ? -1 : ((s.time > stop_time) ? 1 : 0) }
	end

	def draw_grid(x_count, y_count)
		@grid_display_lists ||= {}
		@grid_display_lists[x_count] ||= {}

		@grid_display_lists[x_count][y_count] ||= GL.RenderToList {
			GL.Begin(GL::LINES)
				(-0.5).step(0.5, 1.0 / x_count) { |x| GL.Vertex(x, 0.5) ; GL.Vertex(x, -0.5) }
				(-0.5).step(0.5, 1.0 / y_count) { |y| GL.Vertex(0.5, y) ; GL.Vertex(-0.5, y) }
			GL.End
		}
		GL.CallList(@grid_display_lists[x_count][y_count])
	end

	def draw(percentage, range)
		percent_begin = percentage - range/2.0
		percent_end = percent_begin + range

		index_range = time_range_to_index_range(percent_begin, percent_end)

		index_range = ((index_range.first-1)...(index_range.last)) if index_range.first > 0
		index_range = ((index_range.first)...(index_range.last+1)) if index_range.last < @samples.size-1

		current_index = time_to_sample_index(percentage)

		with_unit_scale {
			#
			# Background
			#
			with_color([@grab ? 0.5 : 0.2, @grab ? 0.5 : 0.3, @grab ? 0.5 : 0.2]) {
				unit_square
			}

			# respect zoom and progress
			scale(1.0 / range, 1.0) {
				translate(-percentage, 0.0) {
					#
					# Samples
					#
					translate(0.0, -0.5) {
						with_color([1.0,1.0,1.0]) {

							x = 0.0
							on = @samples[index_range.first].value rescue false

							GL.LineWidth(16.0)

							s = nil
							for i in index_range.first..index_range.last-1
								s = @samples[i]
								s2 = @samples[i+1]

								if s.value
									if s.time < percentage and (s2.nil? or s2.time > percentage)
										GL.Color(1,1,0)
									else
										GL.Color(1,1,1)
									end

									GL.Begin(GL::LINE)
									GL.Vertex(s.time, 0.5)
									GL.Vertex(s2 ? s2.time : percentage, 0.5)
									GL.End
								end
							end

							GL.LineWidth(1.0)
						}
					}
				}
			}
		}
	end

	def time_to_sample_index(time)
		i = @samples.bsearch_lower_boundary { |s| s.time <=> time }

		# exact match?
		return i if @samples[i] and @samples[i].time == time

		return i-1 # from boundary to index (see bsearch.png)
	end

	def add_sample(time, value)
		if @samples.empty?
			# The easiest case, just tack it on the end
			@samples << Sample.new(time, value)

		elsif time > @samples.last.time
			# Tack on unless it would be redundant
			@samples << Sample.new(time, value) unless @samples.last.value == value

		elsif time == @samples.last.time
			# Update previous entry with latest value
			@samples.last.value = value

		else
			# And now the hardest case, updating somewhere in the middle
			i = time_to_sample_index(time)

			if @previous_update_time
				previous_i = time_to_sample_index(@previous_update_time)

				if previous_i < i
					# overwrite entries with a new value, NOT including previous_i
					@samples[(previous_i + 1), (i - previous_i)] = Sample.new(time, value) #, Sample.new(time, @samples[i].value)]

				# otherwise same sample. new value?
				elsif @samples[previous_i].value != value
					# insert after previous_i
					@samples.insert(previous_i+1, Sample.new(time, value))
				end
			else
				@samples.insert(i+1, Sample.new(time, value)) unless @samples[i] and @samples[i].value == value
			end

			@previous_update_time = time
		end
	end

	def button_down(time, number, x, y)
		return unless within?(x,y)

		@grab = true

		if y >= @y #+ @height/2.0
			@last_y = true
			add_sample(time, @last_y)
		else #if y < @y #- @height/2.0
			@last_y = false
			add_sample(time, @last_y)
		end

#		@last_y = true
#		add_sample(time, @last_y)
	end

	def cursor_move(time, x, y)
		if @grab
			if y >= @y #+ @height/2.0
				@last_y = true
				add_sample(time, @last_y)
			else #if y < @y #- @height/2.0
				@last_y = false
				add_sample(time, @last_y)
			end
		end
	end

	def button_up(time, number, x, y)
		add_sample(time, false)
		@grab = false
	end

	def update(time)
		if @grab
			add_sample(time, @last_y)
		else
			@previous_update_time = nil
		end
	end
end
